__code__='Gas1];01J[&DiU]a)OI^ge]E+<4!>hArtc3(5?673c5:q;,!j))odTZ.,%=g7[DQQ3H6n?;`_f(ItYSQcd43BJgY^VX"c1>qekSE\\,1Tclhg*gR/:-^q#en>0/@b]7K>ZI!!INMCl\\uA$\'54KJ/?\\l5oP819L&ui!S@N[!,MZg!k&-U!%n;d"*bUN*+0$@?,/)NH,+",pAciH^(9u>i\'kEl!C$Xu"F(gRA&\'V5EpgK[>k``["MYSY"eu:C->+GQO:=0)6+lsp2P+(@N7`!_:H6l:j_trhcDRMk#j2RPlN4Up=VX56NMln5X5fIqOi]hCNhll(%75+OCs?Ra!Bfe3J3s>R?CIMf%V_:=694-lJJRp==_4QA3&Dbi=1\\fWk)fY+<1P>D6#joZT)7i7*Y\'6phLG%I1!>8fW<98&\\WG,QND)ctG-\'-7A\\;$#Ubp-fp;kl[rWLcGSGmM-/>lYT?f#tp\\1:q\\EY\\:[>J5lRik.5\'9Oi1NK@7grXtAfr&8ff0*DooXDO(l#bP.3./:G9Lqu/Ro8^-KHhg[?eio:^&VRVM+c7H9[`Yf30Ak@s*Hc\'9sAj;G]WjERYg?[UZTcN6<@)l+mP?aiQ.3])+SSl^Jm:`&=g["7qn%5URN.Xh`9ZO;1QijO2=8.;k\'\'L`<lr;YYRRN+<_DF@L0+\\dpO@;HAEMK-U29Br7I_($DY\\X8;rRQjAD\\T]/<R.#s1prSo.XgZpjqUt\\&tbl3"\\KN4(G#G@d/iq%jS6`!m;Rluc6b$r$$,9.i7_R\\G3.kgRl\'Qi](=J.gY_M_PG3DUnhAHh,sb32Xo?a%5)[F.E^IZ8GdIAP\\NO.09]%?]iJn=jUL_")+DlJ+7CV\\trBFQ5$/Oo3.$e4aRKMrZ+GqIcd*G.m5"^%pSieU96Q&WuU^ugi;&WLfH7`U?=Nt$%/t(teXJk<VH6Z_tYg1>@P%9[moL?2rUJdKsb[eum)l$6Y/S7^-^8R&?`Uo.j>W2OQEgN3@W`&$de_u;"&J^JX.MmIsLgu%:O=eI()j-HH]G.uSfMa?-.d6o^;RX;$f%9X9Sq"R&MY#(]Ypc\'=HnQD]U1sN5?4^Zu~>'



__code__='GatU0D/[on&>$QoP(Q(,!=FNmGmQO3EX1R"#U.^k#VV?^7!CoMUQsqc2Mm::2S4=X71,YPlBhLJrjC%QD+HoA\'#!$?YBOXAH"PV`+V_\\:RQo7Xs,b$S0I06M&tjdL!:Ko:*sthr=1"@m"f_VpUC(**X91?\\%foE+:\'3BT%g*jRr!)X=i%`"b^uKsOJ0t?[!)EZf!^$qe/`7M^Z[gp6Sc_(cpc[OtTcADbAZUZ:SqB6W0W`*p%][>R<"2eIhGlkSq&<@Ae_h1994g10!Z,7k\\kFh\\6"2]&j/^Q)Q/e>YTj/!?)27il^p"1E6&RV;.Z%=`]<)"!dii*p^o=eCYQ0#,WpH8^QnI#`\\PmX8p*u.e..p-1!frs!/seRbcQ":r0,T28s0VNJl!]LDllcu\\ap-F3>Gq.h+?%C0RUVW6T\'NZ(okWt@??]6nQ>)E(L1Q%GUr<?r@BPd?Z\'nc8>eA\\\'h-U#Y@s$?P[Q7n;`MPFt94MdN]?WO^fU8.:F3t].@Rg!19]#@=,T%uf[nZTP8oYjB.kC8?7.mHNkcZ;[05o%<BF#F/08"gYVPid]/Ci;Zf&V__@NZm?!uXV&8rQ*)Bqc8ERBH#-gU&lNfBMYgm\'O1TN.\\N]R#]lUpY>@6S"<:-5s,WnPU809;GFGfPK,B%Fq;/o=>I#?VOl`P];aBAba(Z2B[nt\\T@K:8Qd$g[HkQ4_dY><d1mM5VXSlP2m$Wb:""FjG"3BM00oM&d@KXd/8#=dL.D&2k1)oOu1GGi"RkOO&;0jj$^N#gaSK#%Vh77g?3-o"F%krL%=J_3(cA=1PehpNB&WQS`,;>>QMG+4D(<bKEf/#3fVOK\'jMNWrSj8#CXjRC;r..;,kjEP#*OXpDA\\&;&&H4\\%&$O>!OF.Ye1<IB-HS_nu*a#^N<nn"\'*0!VRiDunP5DXes.aT@h/CI,)g/A-+bJ-mgq#h0@8\\Keb$_Q%AXCu&k]N;N%$)T*u1^>J*@Vr=Y\'o$q=-@&KEQ&pp40CQ:<>[4$$9f2H6.o)0\\kM#`dBL\'bE)\'Nm:KNm<rSgO7+jrr6$4$iGJ1:*-WLo7O+C\'9-]\\/K4ZZ5?!s`NQT!kr-j"~>'



#copyright ReportLab Inc. 2000-2019
#see license.txt for license details
Version='3.5.27'

__version__=Version
__date__='20191001'
import sys

if sys.version_info[0:2]!=(2, 7) and sys.version_info<(3, 5):
    raise ImportError("""rlextra requires Python 2.7+ or 3.5+; 3.0-3.4 are not supported.
If you want to try with other python versions edit line 10 of rlextra/__init__
to remove this error.""")

import reportlab
if reportlab.Version < Version:
	raise ImportError("""rlextra version %s requires at least reportlab %s, and you have reportlab %s; it is strongly recommended that you use the same versions""" % (Version, Version.rsplit('.',1)[0]+'.0', reportlab.Version))



''' Tests for graphics. '''
import unittest,os,sys
from reportlab.lib.logger import warnOnce

from legendedpie import LegendedPieDrawing
from scatterplot import ScatterPlotDrawing
from slidebox import SlideBoxDrawing
from stackedBar import CylinderBarStacked
TESTS=(
        LegendedPieDrawing,
        ScatterPlotDrawing,
        SlideBoxDrawing,
        CylinderBarStacked,
        )

class devnull:
    def write(*a, **kw):
        pass

class TestCase(unittest.TestCase):

    def setUp(self):
        self.stdout = sys.stdout
        sys.stdout = devnull()
        self.olddir = os.getcwd()
        os.chdir(os.path.abspath(os.path.dirname(__file__)))

    def tearDown(self):
        os.chdir(self.olddir)
        sys.stdout = self.stdout

def makeSuite():
    class NullTest(unittest.TestCase):
        pass

    try:
        from reportlab.graphics import renderPM
    except ImportError:
        warnOnce('''Could not import _renderPM.
Therefore the rlextra.examples.graphics tests won't be run.
See http://www.reportlab.org/rl_addons.html''')
        return unittest.makeSuite(NullTest)
    canv = renderPM.PMCanvas(1,1)
    try:
        canv.setFont('Times-Roman', 10)
    except renderPM.RenderPMError:
        warnOnce('''Could not find the standard PFB fonts in reportlab/fonts.
Therefore the rlextra.examples.graphics tests won't be run.
The required fonts may be downloaded from http://www.reportlab.com/ftp/pfbfer.zip
and unzipped into reportlab/fonts.''')
        return unittest.makeSuite(NullTest)
    del canv, renderPM

    # Add tests to class before making the suite
    for test in TESTS:
        def func(self,t=test):
            t().go()
        setattr(TestCase, 'test_%s' % test.__name__, func)
    return unittest.makeSuite(TestCase)

if __name__ == "__main__":
    unittest.TextTestRunner().run(makeSuite())



#copyright ReportLab Europe Limited. 2000-2016
#see license.txt for license details
__version__='3.3.0'
from rlextra.ers.webapp import WebApp
from reportlab.lib.colors import toColor
from rlextra.graphics.quickchart import quickChart, getTestDrawing

class WebChartApp(WebApp):
    "Web test harness / demonstrator for our charts"
    def __init__(self,**kw):
        WebApp.__init__(self,**kw)
        self.viewDir = '.'
        self.defaultAction = 'main'
        
    def action_testImage(self, request, response):
        myChart = getTestDrawing()
        bitmapBytes = myChart.asString('png')
        response.setContentType('image/png')
        response.write(bitmapBytes)

    def action_getLogo(self, request, response):
        "Return reportlab logo"
        bitmapBytes = open('corplogo.gif','rb').read()
        response.setContentType('image/png')
        response.write(bitmapBytes)
        
    def action_image(self, request, response):
        "Generate PNG chart according to input parameters"


        # extract any chart-related parameters, converting
        # from strings to correct data types.  Keep this
        # explicit instead of being clever and dynamic
        params = request.params
        kw = {}
        def getColor(name,kw=kw,params=params):
            color = params.get(name,'None')
            kw[name] = color.upper()!='NONE' and toColor(color) or None

        def getParam(name,kw=kw,params=params):
            if name in params: kw[name] = params[name]

        kw['showBoundaries'] = 'showBoundaries' in params
        kw['chartType'] = params.get('chartType','bar')
        kw['seriesRelation'] = params.get('seriesRelation', 'sidebyside')
        kw['width'] = params.get('width','400')
        kw['height'] = params.get('height','300')

        getParam('data')
        kw['textData'] = params.get('textData',None)
        getParam('categoryNames')
        getParam('seriesNames')

        #chart colours
        if 'chartColors' in params:
            kw['chartColors'] = params['chartColors']
            chartColorText = params['chartColors']
            if chartColorText == 'None':
                chartColors = None
            else:
                chartColors = chartColorText
        else:
            chartColors = None
        kw['chartColors'] = chartColors

        #main title
        getParam('titleText')
        getParam('titleFontName')
        getParam('titleFontSize')
        getParam('titleFontColor')

        #x axis title
        getParam('xTitleText')
        getParam('xTitleFontName')
        getParam('xTitleFontSize')
        getParam('xTitleFontColor')

        #y axis title
        getParam('yTitleText')
        getParam('yTitleFontName')
        getParam('yTitleFontSize')
        getParam('yTitleFontColor')

        #x axis 
        kw['xAxisVisible'] = 'xAxisVisible' in params
        getParam('xAxisFontName')
        getParam('xAxisFontSize')
        getParam('xAxisFontColor')
        getParam('xAxisLabelAngle')

        #y axis 
        kw['yAxisVisible'] = 'yAxisVisible' in params
        getParam('yAxisFontName')
        getParam('yAxisFontSize')
        getParam('yAxisFontColor')
        getParam('yAxisLabelAngle')

        #data labels
        getParam('dataLabelsType')
        getParam('dataLabelsFontName')
        getParam('dataLabelsFontSize')
        getParam('dataLabelsFontColor')
        getParam('dataLabelsAlignment')

        # plot area
        kw['xAxisGridLines'] = 'xAxisGridLines' in params
        kw['yAxisGridLines'] = 'yAxisGridLines' in params
        
        if 'plotColor' in request.params:
            kw['plotColor'] =  request.params['plotColor']
        else:
            kw['plotColor'] =  None

        #if they give us a background colour, use it.
        getColor('bgColor')

        #if they give us a background stroke colour, use it.
        getColor('bgStrokeColor')

        #markers for lines eg, in scatter_lines_markers
        if 'markerType' in params:
            mk = params['markerType']
            if mk == 'None':
                mk = None
            kw['markerType'] = mk

        if 'markerSize' in params:
            ms = params['markerSize']
            if ms == 'None':
                ms = None
            else:
                ms = int(ms)
            kw['markerSize'] = ms

        #legend properties
        if 'legendPos' in params:
            lpos = params['legendPos']
            if lpos == 'None':
                lpos = None
            kw['legendPos'] = lpos

        if 'legendFontName' in params:
            lfont = params['legendFontName']
            if lfont == 'None':
                lfont = None
            kw['legendFontName'] = lfont

        if 'legendFontSize' in params:
            lsize = params['legendFontSize']
            if lsize == 'None':
                lsize = None
            kw['legendFontSize'] = lsize

        if 'legendFontColor' in params:
            lcol = params['legendFontColor']
            if lcol == 'None':
                lcol = None
            kw['legendFontColor'] = lcol

        myChart = quickChart(**kw)
        bitmapBytes = myChart.asString('gif')
        response.setContentType('image/gif')
        response.setHeader('Cache-Control', 'no-cache')
        response.write(bitmapBytes)

if __name__=='__main__':
    WebChartApp().handleCommandLine()



#!/usr/bin/env python

"""
xcharts.py - a proof-of-concept XML interface to Diagra

This is a demonstration prepared for a customer and may not be fully robust
or secure yet.

"""
import os, sys
import optparse


def run(inFileName, options):
    print("Processing", infilename)

def parseCommandLine():
    return None, None



sample = """
<drawings>
  <drawing module="gridlineplot" class="GridLinePlotDrawing">
  
  </drawing>
</drawings>
"""


if __name__=='__main__':
    options, args = parseCommandLine
    if not args:
        print("usage: xcharts.py infile.xml [..options...]")
    else:
        run(args[0], options)
        




# vim: sw=4:expandtab:foldmethod=marker
#
# Copyright (c) 2006, Mathieu Fenniak
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# * The name of the author may not be used to endorse or promote products
# derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


"""
Implementation of generic PDF objects (dictionary, number, string, and so on)
"""
__author__ = "Mathieu Fenniak"
__author_email__ = "biziqe@mathieu.fenniak.net"

import re
from .utils import readNonWhitespace, RC4_encrypt, skipOverComment
from .utils import b_, u_, chr_, ord_
from .utils import PdfStreamError
import warnings
from . import filters
from . import utils
import decimal
import codecs
#import debugging

ObjectPrefix = b_('/<[tf(n%')
NumberSigns = b_('+-')
IndirectPattern = re.compile(b_(r"(\d+)\s+(\d+)\s+R[^a-zA-Z]"))


def readObject(stream, pdf):
	tok = stream.read(1)
	stream.seek(-1, 1) # reset to start
	idx = ObjectPrefix.find(tok)
	if idx == 0:
		# name object
		return NameObject.readFromStream(stream, pdf)
	elif idx == 1:
		# hexadecimal string OR dictionary
		peek = stream.read(2)
		stream.seek(-2, 1) # reset to start
		if peek == b_('<<'):
			return DictionaryObject.readFromStream(stream, pdf)
		else:
			return readHexStringFromStream(stream)
	elif idx == 2:
		# array object
		return ArrayObject.readFromStream(stream, pdf)
	elif idx == 3 or idx == 4:
		# boolean object
		return BooleanObject.readFromStream(stream)
	elif idx == 5:
		# string object
		return readStringFromStream(stream)
	elif idx == 6:
		# null object
		return NullObject.readFromStream(stream)
	elif idx == 7:
		# comment
		while tok not in (b_('\r'), b_('\n')):
			tok = stream.read(1)
		tok = readNonWhitespace(stream)
		stream.seek(-1, 1)
		return readObject(stream, pdf)
	else:
		# number object OR indirect reference
		if tok in NumberSigns:
			# number
			return NumberObject.readFromStream(stream)
		peek = stream.read(20)
		stream.seek(-len(peek), 1) # reset to start
		if IndirectPattern.match(peek) != None:
			return IndirectObject.readFromStream(stream, pdf)
		else:
			return NumberObject.readFromStream(stream)


class PdfObject(object):
	def getObject(self):
		"""Resolves indirect references."""
		return self


class NullObject(PdfObject):
	def writeToStream(self, stream, encryption_key):
		stream.write(b_("null"))

	def readFromStream(stream):
		nulltxt = stream.read(4)
		if nulltxt != b_("null"):
			raise utils.PdfReadError("Could not read Null object")
		return NullObject()
	readFromStream = staticmethod(readFromStream)


class BooleanObject(PdfObject):
	def __init__(self, value):
		self.value = value

	def writeToStream(self, stream, encryption_key):
		if self.value:
			stream.write(b_("true"))
		else:
			stream.write(b_("false"))

	def readFromStream(stream):
		word = stream.read(4)
		if word == b_("true"):
			return BooleanObject(True)
		elif word == b_("fals"):
			stream.read(1)
			return BooleanObject(False)
		else:
			raise utils.PdfReadError('Could not read Boolean object')
	readFromStream = staticmethod(readFromStream)


class ArrayObject(list, PdfObject):
	def writeToStream(self, stream, encryption_key):
		stream.write(b_("["))
		for data in self:
			stream.write(b_(" "))
			data.writeToStream(stream, encryption_key)
		stream.write(b_(" ]"))

	def readFromStream(stream, pdf):
		arr = ArrayObject()
		tmp = stream.read(1)
		if tmp != b_("["):
			raise utils.PdfReadError("Could not read array")
		while True:
			# skip leading whitespace
			tok = stream.read(1)
			while tok.isspace():
				tok = stream.read(1)
			stream.seek(-1, 1)
			# check for array ending
			peekahead = stream.read(1)
			if peekahead == b_("]"):
				break
			stream.seek(-1, 1)
			# read and append obj
			arr.append(readObject(stream, pdf))
		return arr
	readFromStream = staticmethod(readFromStream)


class IndirectObject(PdfObject):
	def __init__(self, idnum, generation, pdf):
		self.idnum = idnum
		self.generation = generation
		self.pdf = pdf

	def getObject(self):
		return self.pdf.getObject(self).getObject()

	def __repr__(self):
		return "IndirectObject(%r, %r)" % (self.idnum, self.generation)

	def __eq__(self, other):
		return (
			other != None and
			isinstance(other, IndirectObject) and
			self.idnum == other.idnum and
			self.generation == other.generation and
			self.pdf is other.pdf
			)

	def __ne__(self, other):
		return not self.__eq__(other)

	def writeToStream(self, stream, encryption_key):
		stream.write(b_("%s %s R" % (self.idnum, self.generation)))

	def readFromStream(stream, pdf):
		idnum = b_("")
		while True:
			tok = stream.read(1)
			if not tok:
				# stream has truncated prematurely
				raise PdfStreamError("Stream has ended unexpectedly")
			if tok.isspace():
				break
			idnum += tok
		generation = b_("")
		while True:
			tok = stream.read(1)
			if not tok:
				# stream has truncated prematurely
				raise PdfStreamError("Stream has ended unexpectedly")
			if tok.isspace():
				if not generation:
					continue
				break
			generation += tok
		r = readNonWhitespace(stream)
		if r != b_("R"):
			raise utils.PdfReadError("Error reading indirect object reference at byte %s" % utils.hexStr(stream.tell()))
		return IndirectObject(int(idnum), int(generation), pdf)
	readFromStream = staticmethod(readFromStream)


class FloatObject(decimal.Decimal, PdfObject):
	def __new__(cls, value="0", context=None):
		try:
			return decimal.Decimal.__new__(cls, utils.str_(value), context)
		except:
			return decimal.Decimal.__new__(cls, str(value))

	def __repr__(self):
		if self == self.to_integral():
			return str(self.quantize(decimal.Decimal(1)))
		else:
			# XXX: this adds useless extraneous zeros.
			return "%.5f" % self

	def as_numeric(self):
		return float(b_(repr(self)))

	def writeToStream(self, stream, encryption_key):
		stream.write(b_(repr(self)))


class NumberObject(int, PdfObject):
	NumberPattern = re.compile(b_('[^+-.0-9]'))
	ByteDot = b_(".")

	def __new__(cls, value):
		return int.__new__(cls, value)

	def as_numeric(self):
		return int(b_(repr(self)))

	def writeToStream(self, stream, encryption_key):
		stream.write(b_(repr(self)))

	def readFromStream(stream):
		num = utils.readUntilRegex(stream, NumberObject.NumberPattern)
		if num.find(NumberObject.ByteDot) != -1:
			return FloatObject(num)
		else:
			return NumberObject(num)
	readFromStream = staticmethod(readFromStream)


##
# Given a string (either a "str" or "unicode"), create a ByteStringObject or a
# TextStringObject to represent the string.
def createStringObject(string):
	if isinstance(string, utils.string_type):
		return TextStringObject(string)
	elif isinstance(string, utils.bytes_type):
		try:
			if string.startswith(codecs.BOM_UTF16_BE):
				retval = TextStringObject(string.decode("utf-16"))
				retval.autodetect_utf16 = True
				return retval
			else:
				# This is probably a big performance hit here, but we need to
				# convert string objects into the text/unicode-aware version if
				# possible... and the only way to check if that's possible is
				# to try.  Some strings are strings, some are just byte arrays.
				retval = TextStringObject(decode_pdfdocencoding(string))
				retval.autodetect_pdfdocencoding = True
				return retval
		except UnicodeDecodeError:
			return ByteStringObject(string)
	else:
		raise TypeError("createStringObject should have str or unicode arg")


def readHexStringFromStream(stream):
	stream.read(1)
	txt = ""
	x = b_("")
	while True:
		tok = readNonWhitespace(stream)
		if not tok:
			# stream has truncated prematurely
			raise PdfStreamError("Stream has ended unexpectedly")
		if tok == b_(">"):
			break
		x += tok
		if len(x) == 2:
			txt += chr(int(x, base=16))
			x = b_("")
	if len(x) == 1:
		x += b_("0")
	if len(x) == 2:
		txt += chr(int(x, base=16))
	return createStringObject(b_(txt))


def readStringFromStream(stream):
	tok = stream.read(1)
	parens = 1
	txt = b_("")
	while True:
		tok = stream.read(1)
		if not tok:
			# stream has truncated prematurely
			raise PdfStreamError("Stream has ended unexpectedly")
		if tok == b_("("):
			parens += 1
		elif tok == b_(")"):
			parens -= 1
			if parens == 0:
				break
		elif tok == b_("\\"):
			tok = stream.read(1)
			if tok == b_("n"):
				tok = b_("\n")
			elif tok == b_("r"):
				tok = b_("\r")
			elif tok == b_("t"):
				tok = b_("\t")
			elif tok == b_("b"):
				tok = b_("\b")
			elif tok == b_("f"):
				tok = b_("\f")
			elif tok == b_("("):
				tok = b_("(")
			elif tok == b_(")"):
				tok = b_(")")
			elif tok == b_("\\"):
				tok = b_("\\")
			elif tok in (b_(" "), b_("/"), b_("%"), b_("<"), b_(">"), b_("["), b_("]")):
				# odd/unnessecary escape sequences we have encountered
				tok = b_(tok)
			elif tok.isdigit():
				# "The number ddd may consist of one, two, or three
				# octal digits; high-order overflow shall be ignored.
				# Three octal digits shall be used, with leading zeros
				# as needed, if the next character of the string is also
				# a digit." (PDF reference 7.3.4.2, p 16)
				for i in range(2):
					ntok = stream.read(1)
					if ntok.isdigit():
						tok += ntok
					else:
						break
				tok = b_(chr(int(tok, base=8)))
			elif tok in b_("\n\r"):
				# This case is	hit when a backslash followed by a line
				# break occurs.  If it's a multi-char EOL, consume the
				# second character:
				tok = stream.read(1)
				if not tok in b_("\n\r"):
					stream.seek(-1, 1)
				# Then don't add anything to the actual string, since this
				# line break was escaped:
				tok = b_('')
			else:
				raise utils.PdfReadError("Unexpected escaped string")
		txt += tok
	return createStringObject(txt)


##
# Represents a string object where the text encoding could not be determined.
# This occurs quite often, as the PDF spec doesn't provide an alternate way to
# represent strings -- for example, the encryption data stored in files (like
# /O) is clearly not text, but is still stored in a "String" object.
class ByteStringObject(utils.bytes_type, PdfObject):

	##
	# For compatibility with TextStringObject.original_bytes.  This method
	# returns self.
	original_bytes = property(lambda self: self)

	def writeToStream(self, stream, encryption_key):
		bytearr = self
		if encryption_key:
			bytearr = RC4_encrypt(encryption_key, bytearr)
		stream.write(b_("<"))
		stream.write(utils.hexencode(bytearr))
		stream.write(b_(">"))


##
# Represents a string object that has been decoded into a real unicode string.
# If read from a PDF document, this string appeared to match the
# PDFDocEncoding, or contained a UTF-16BE BOM mark to cause UTF-16 decoding to
# occur.
class TextStringObject(utils.string_type, PdfObject):
	autodetect_pdfdocencoding = False
	autodetect_utf16 = False

	##
	# It is occasionally possible that a text string object gets created where
	# a byte string object was expected due to the autodetection mechanism --
	# if that occurs, this "original_bytes" property can be used to
	# back-calculate what the original encoded bytes were.
	original_bytes = property(lambda self: self.get_original_bytes())

	def get_original_bytes(self):
		# We're a text string object, but the library is trying to get our raw
		# bytes.  This can happen if we auto-detected this string as text, but
		# we were wrong.  It's pretty common.  Return the original bytes that
		# would have been used to create this object, based upon the autodetect
		# method.
		if self.autodetect_utf16:
			return codecs.BOM_UTF16_BE + self.encode("utf-16be")
		elif self.autodetect_pdfdocencoding:
			return encode_pdfdocencoding(self)
		else:
			raise Exception("no information about original bytes")

	def writeToStream(self, stream, encryption_key):
		# Try to write the string out as a PDFDocEncoding encoded string.  It's
		# nicer to look at in the PDF file.  Sadly, we take a performance hit
		# here for trying...
		try:
			bytearr = encode_pdfdocencoding(self)
		except UnicodeEncodeError:
			bytearr = codecs.BOM_UTF16_BE + self.encode("utf-16be")
		if encryption_key:
			bytearr = RC4_encrypt(encryption_key, bytearr)
			obj = ByteStringObject(bytearr)
			obj.writeToStream(stream, None)
		else:
			stream.write(b_("("))
			for c in bytearr:
				if not chr_(c).isalnum() and c != b_(' '):
					stream.write(b_("\\%03o" % ord_(c)))
				else:
					stream.write(b_(chr_(c)))
			stream.write(b_(")"))


class NameObject(str, PdfObject):
	delimiterPattern = re.compile(b_(r"\s+|[()<>[\]{}/%]"))
	surfix = b_("/")

	def writeToStream(self, stream, encryption_key):
		stream.write(b_(self))

	def readFromStream(stream, pdf):
		debug = False
		if debug: print((stream.tell()))
		name = stream.read(1)
		if name != NameObject.surfix:
			raise utils.PdfReadError("name read error")
		name += utils.readUntilRegex(stream, NameObject.delimiterPattern)
		if debug: print(name)
		try:
			return NameObject(name.decode('utf-8'))
		except UnicodeDecodeError as e:
			# Name objects should represent irregular characters
			# with a '#' followed by the symbol's hex number
			if not pdf.strict:
				warnings.warn("Illegal character in Name Object", utils.PdfReadWarning)
				return NameObject(name)
			else:
				raise utils.PdfReadError("Illegal character in Name Object")

	readFromStream = staticmethod(readFromStream)


class DictionaryObject(dict, PdfObject):
	def raw_get(self, key):
		return dict.__getitem__(self, key)

	def __setitem__(self, key, value):
		if not isinstance(key, PdfObject):
			raise ValueError("key must be PdfObject")
		if not isinstance(value, PdfObject):
			raise ValueError("value must be PdfObject")
		return dict.__setitem__(self, key, value)

	def setdefault(self, key, value=None):
		if not isinstance(key, PdfObject):
			raise ValueError("key must be PdfObject")
		if not isinstance(value, PdfObject):
			raise ValueError("value must be PdfObject")
		return dict.setdefault(self, key, value)

	def __getitem__(self, key):
		return dict.__getitem__(self, key).getObject()

	##
	# Retrieves XMP (Extensible Metadata Platform) data relevant to the
	# this object, if available.
	# <p>
	# Stability: Added in v1.12, will exist for all future v1.x releases.
	# @return Returns a {@link #xmp.XmpInformation XmlInformation} instance
	# that can be used to access XMP metadata from the document.  Can also
	# return None if no metadata was found on the document root.
	def getXmpMetadata(self):
		metadata = self.get("/Metadata", None)
		if metadata == None:
			return None
		metadata = metadata.getObject()
		from . import xmp
		if not isinstance(metadata, xmp.XmpInformation):
			metadata = xmp.XmpInformation(metadata)
			self[NameObject("/Metadata")] = metadata
		return metadata

	##
	# Read-only property that accesses the {@link
	# #DictionaryObject.getXmpData getXmpData} function.
	# <p>
	# Stability: Added in v1.12, will exist for all future v1.x releases.
	xmpMetadata = property(lambda self: self.getXmpMetadata(), None, None)

	def writeToStream(self, stream, encryption_key):
		stream.write(b_("<<\n"))
		for key, value in list(self.items()):
			key.writeToStream(stream, encryption_key)
			stream.write(b_(" "))
			value.writeToStream(stream, encryption_key)
			stream.write(b_("\n"))
		stream.write(b_(">>"))

	def readFromStream(stream, pdf):
		debug = False
		tmp = stream.read(2)
		if tmp != b_("<<"):
			raise utils.PdfReadError("Dictionary read error at byte %s: stream must begin with '<<'" % utils.hexStr(stream.tell()))
		data = {}
		while True:
			tok = readNonWhitespace(stream)
			if tok == b_('\x00'):
				continue
			elif tok == b_('%'):
				stream.seek(-1, 1)
				skipOverComment(stream)
				continue
			if not tok:
				# stream has truncated prematurely
				raise PdfStreamError("Stream has ended unexpectedly")

			if debug: print(("Tok:", tok))
			if tok == b_(">"):
				stream.read(1)
				break
			stream.seek(-1, 1)
			key = readObject(stream, pdf)
			tok = readNonWhitespace(stream)
			stream.seek(-1, 1)
			value = readObject(stream, pdf)
			if not data.get(key):
				data[key] = value
			elif pdf.strict:
				# multiple definitions of key not permitted
				raise utils.PdfReadError("Multiple definitions in dictionary at byte %s for key %s" \
										   % (utils.hexStr(stream.tell()), key))
			else:
				warnings.warn("Multiple definitions in dictionary at byte %s for key %s" \
										   % (utils.hexStr(stream.tell()), key), utils.PdfReadWarning)

		pos = stream.tell()
		s = readNonWhitespace(stream)
		if s == b_('s') and stream.read(5) == b_('tream'):
			eol = stream.read(1)
			# odd PDF file output has spaces after 'stream' keyword but before EOL.
			# patch provided by Danial Sandler
			while eol == b_(' '):
				eol = stream.read(1)
			assert eol in (b_("\n"), b_("\r"))
			if eol == b_("\r"):
				# read \n after
				if stream.read(1)  != b_('\n'):
					stream.seek(-1, 1)
			# this is a stream object, not a dictionary
			assert "/Length" in data
			length = data["/Length"]
			if debug: print(data)
			if isinstance(length, IndirectObject):
				t = stream.tell()
				length = pdf.getObject(length)
				stream.seek(t, 0)
			data["__streamdata__"] = stream.read(length)
			if debug: print("here")
			#if debug: print(binascii.hexlify(data["__streamdata__"]))
			e = readNonWhitespace(stream)
			ndstream = stream.read(8)
			if (e + ndstream) != b_("endstream"):
				# (sigh) - the odd PDF file has a length that is too long, so
				# we need to read backwards to find the "endstream" ending.
				# ReportLab (unknown version) generates files with this bug,
				# and Python users into PDF files tend to be our audience.
				# we need to do this to correct the streamdata and chop off
				# an extra character.
				pos = stream.tell()
				stream.seek(-10, 1)
				end = stream.read(9)
				if end == b_("endstream"):
					# we found it by looking back one character further.
					data["__streamdata__"] = data["__streamdata__"][:-1]
				else:
					if debug: print(("E", e, ndstream, debugging.toHex(end)))
					stream.seek(pos, 0)
					raise utils.PdfReadError("Unable to find 'endstream' marker after stream at byte %s." % utils.hexStr(stream.tell()))
		else:
			stream.seek(pos, 0)
		if "__streamdata__" in data:
			return StreamObject.initializeFromDictionary(data)
		else:
			retval = DictionaryObject()
			retval.update(data)
			return retval
	readFromStream = staticmethod(readFromStream)


class TreeObject(DictionaryObject):
	def __init__(self):
		DictionaryObject.__init__(self)

	def hasChildren(self):
		return '/First' in self

	def __iter__(self):
		return self.children()

	def children(self):
		if not self.hasChildren():
			raise StopIteration

		child = self['/First']
		while True:
			yield child
			if child == self['/Last']:
				raise StopIteration
			child = child['/Next']

	def addChild(self, child, pdf):
		childObj = child.getObject()
		child = pdf.getReference(childObj)
		assert isinstance(child, IndirectObject)

		if '/First' not in self:
			self[NameObject('/First')] = child
			self[NameObject('/Count')] = NumberObject(0)
			prev = None
		else:
			prev = self['/Last']

		self[NameObject('/Last')] = child
		self[NameObject('/Count')] = NumberObject(self[NameObject('/Count')] + 1)

		if prev:
			prevRef = pdf.getReference(prev)
			assert isinstance(prevRef, IndirectObject)
			childObj[NameObject('/Prev')] = prevRef
			prev[NameObject('/Next')] = child

		parentRef = pdf.getReference(self)
		assert isinstance(parentRef, IndirectObject)
		childObj[NameObject('/Parent')] = parentRef

	def removeChild(self, child):
		childObj = child.getObject()

		if NameObject('/Parent') not in childObj:
			raise ValueError("Removed child does not appear to be a tree item")
		elif childObj[NameObject('/Parent')] != self:
			raise ValueError("Removed child is not a member of this tree")

		found = False
		prevRef = None
		prev = None
		curRef = self[NameObject('/First')]
		cur = curRef.getObject()
		lastRef = self[NameObject('/Last')]
		last = lastRef.getObject()
		while cur != None:
			if cur == childObj:
				if prev == None:
					if NameObject('/Next') in cur:
						# Removing first tree node
						nextRef = cur[NameObject('/Next')]
						next = nextRef.getObject()
						del next[NameObject('/Prev')]
						self[NameObject('/First')] = nextRef
						self[NameObject('/Count')] = self[NameObject('/Count')] - 1

					else:
						# Removing only tree node
						assert self[NameObject('/Count')] == 1
						del self[NameObject('/Count')]
						del self[NameObject('/First')]
						if NameObject('/Last') in self:
							del self[NameObject('/Last')]
				else:
					if NameObject('/Next') in cur:
						# Removing middle tree node
						nextRef = cur[NameObject('/Next')]
						next = nextRef.getObject()
						next[NameObject('/Prev')] = prevRef
						prev[NameObject('/Next')] = nextRef
						self[NameObject('/Count')] = self[NameObject('/Count')] - 1
					else:
						# Removing last tree node
						assert cur == last
						del prev[NameObject('/Next')]
						self[NameObject('/Last')] = prevRef
						self[NameObject('/Count')] = self[NameObject('/Count')] - 1
				found = True
				break

			prevRef = curRef
			prev = cur
			if NameObject('/Next') in cur:
				curRef = cur[NameObject('/Next')]
				cur = curRef.getObject()
			else:
				curRef = None
				cur = None

		if not found:
			raise ValueError("Removal couldn't find item in tree")

		del childObj[NameObject('/Parent')]
		if NameObject('/Next') in childObj:
			del childObj[NameObject('/Next')]
		if NameObject('/Prev') in childObj:
			del childObj[NameObject('/Prev')]

	def emptyTree(self):
		for child in self:
			childObj = child.getObject()
			del childObj[NameObject('/Parent')]
			if NameObject('/Next') in childObj:
				del childObj[NameObject('/Next')]
			if NameObject('/Prev') in childObj:
				del childObj[NameObject('/Prev')]

		if NameObject('/Count') in self:
			del self[NameObject('/Count')]
		if NameObject('/First') in self:
			del self[NameObject('/First')]
		if NameObject('/Last') in self:
			del self[NameObject('/Last')]


class StreamObject(DictionaryObject):
	def __init__(self):
		self._data = None
		self.decodedSelf = None

	def writeToStream(self, stream, encryption_key):
		self[NameObject("/Length")] = NumberObject(len(self._data))
		DictionaryObject.writeToStream(self, stream, encryption_key)
		del self["/Length"]
		stream.write(b_("\nstream\n"))
		data = self._data
		if encryption_key:
			data = RC4_encrypt(encryption_key, data)
		stream.write(data)
		stream.write(b_("\nendstream"))

	def initializeFromDictionary(data):
		if "/Filter" in data:
			retval = EncodedStreamObject()
		else:
			retval = DecodedStreamObject()
		retval._data = data["__streamdata__"]
		del data["__streamdata__"]
		del data["/Length"]
		retval.update(data)
		return retval
	initializeFromDictionary = staticmethod(initializeFromDictionary)

	def flateEncode(self):
		if "/Filter" in self:
			f = self["/Filter"]
			if isinstance(f, ArrayObject):
				f.insert(0, NameObject("/FlateDecode"))
			else:
				newf = ArrayObject()
				newf.append(NameObject("/FlateDecode"))
				newf.append(f)
				f = newf
		else:
			f = NameObject("/FlateDecode")
		retval = EncodedStreamObject()
		retval[NameObject("/Filter")] = f
		retval._data = filters.FlateDecode.encode(self._data)
		return retval


class DecodedStreamObject(StreamObject):
	def getData(self):
		return self._data

	def setData(self, data):
		self._data = data


class EncodedStreamObject(StreamObject):
	def __init__(self):
		self.decodedSelf = None

	def getData(self):
		if self.decodedSelf:
			# cached version of decoded object
			return self.decodedSelf.getData()
		else:
			# create decoded object
			decoded = DecodedStreamObject()

			decoded._data = filters.decodeStreamData(self)
			for key, value in list(self.items()):
				if not key in ("/Length", "/Filter", "/DecodeParms"):
					decoded[key] = value
			self.decodedSelf = decoded
			return decoded._data

	def setData(self, data):
		raise utils.PdfReadError("Creating EncodedStreamObject is not currently supported")


class RectangleObject(ArrayObject):
	"""
	This class is used to represent *page boxes* in PyPDF2. These boxes include:

		* :attr:`artBox <PyPDF2.pdf.PageObject.artBox>`
		* :attr:`bleedBox <PyPDF2.pdf.PageObject.bleedBox>`
		* :attr:`cropBox <PyPDF2.pdf.PageObject.cropBox>`
		* :attr:`mediaBox <PyPDF2.pdf.PageObject.mediaBox>`
		* :attr:`trimBox <PyPDF2.pdf.PageObject.trimBox>`
	"""
	def __init__(self, arr):
		# must have four points
		assert len(arr) == 4
		# automatically convert arr[x] into NumberObject(arr[x]) if necessary
		ArrayObject.__init__(self, [self.ensureIsNumber(x) for x in arr])

	def ensureIsNumber(self, value):
		if not isinstance(value, (NumberObject, FloatObject)):
			value = FloatObject(value)
		return value

	def __repr__(self):
		return "RectangleObject(%s)" % repr(list(self))

	def getLowerLeft_x(self):
		return self[0]

	def getLowerLeft_y(self):
		return self[1]

	def getUpperRight_x(self):
		return self[2]

	def getUpperRight_y(self):
		return self[3]

	def getUpperLeft_x(self):
		return self.getLowerLeft_x()

	def getUpperLeft_y(self):
		return self.getUpperRight_y()

	def getLowerRight_x(self):
		return self.getUpperRight_x()

	def getLowerRight_y(self):
		return self.getLowerLeft_y()

	def getLowerLeft(self):
		return self.getLowerLeft_x(), self.getLowerLeft_y()

	def getLowerRight(self):
		return self.getLowerRight_x(), self.getLowerRight_y()

	def getUpperLeft(self):
		return self.getUpperLeft_x(), self.getUpperLeft_y()

	def getUpperRight(self):
		return self.getUpperRight_x(), self.getUpperRight_y()

	def setLowerLeft(self, value):
		self[0], self[1] = [self.ensureIsNumber(x) for x in value]

	def setLowerRight(self, value):
		self[2], self[1] = [self.ensureIsNumber(x) for x in value]

	def setUpperLeft(self, value):
		self[0], self[3] = [self.ensureIsNumber(x) for x in value]

	def setUpperRight(self, value):
		self[2], self[3] = [self.ensureIsNumber(x) for x in value]

	def getWidth(self):
		return self.getUpperRight_x() - self.getLowerLeft_x()

	def getHeight(self):
		return self.getUpperRight_y() - self.getLowerLeft_y()

	lowerLeft = property(getLowerLeft, setLowerLeft, None, None)
	"""
	Property to read and modify the lower left coordinate of this box
	in (x,y) form.
	"""
	lowerRight = property(getLowerRight, setLowerRight, None, None)
	"""
	Property to read and modify the lower right coordinate of this box
	in (x,y) form.
	"""
	upperLeft = property(getUpperLeft, setUpperLeft, None, None)
	"""
	Property to read and modify the upper left coordinate of this box
	in (x,y) form.
	"""
	upperRight = property(getUpperRight, setUpperRight, None, None)
	"""
	Property to read and modify the upper right coordinate of this box
	in (x,y) form.
	"""


class Field(TreeObject):
	"""
	A class representing a field dictionary. This class is accessed through
	:meth:`getFields()<PyPDF2.PdfFileReader.getFields>`
	"""
	def __init__(self, data):
		DictionaryObject.__init__(self)
		attributes = ("/FT", "/Parent", "/Kids", "/T", "/TU", "/TM", "/Ff",
					  "/V", "/DV", "/AA")
		for attr in attributes:
			try:
				self[NameObject(attr)] = data[attr]
			except KeyError:
				pass

	fieldType = property(lambda self: self.get("/FT"))
	"""
	Read-only property accessing the type of this field.
	"""

	parent = property(lambda self: self.get("/Parent"))
	"""
	Read-only property accessing the parent of this field.
	"""

	kids = property(lambda self: self.get("/Kids"))
	"""
	Read-only property accessing the kids of this field.
	"""

	name = property(lambda self: self.get("/T"))
	"""
	Read-only property accessing the name of this field.
	"""

	altName = property(lambda self: self.get("/TU"))
	"""
	Read-only property accessing the alternate name of this field.
	"""

	mappingName = property(lambda self: self.get("/TM"))
	"""
	Read-only property accessing the mapping name of this field. This
	name is used by PyPDF2 as a key in the dictionary returned by
	:meth:`getFields()<PyPDF2.PdfFileReader.getFields>`
	"""

	flags = property(lambda self: self.get("/Ff"))
	"""
	Read-only property accessing the field flags, specifying various
	characteristics of the field (see Table 8.70 of the PDF 1.7 reference).
	"""

	value = property(lambda self: self.get("/V"))
	"""
	Read-only property accessing the value of this field. Format
	varies based on field type.
	"""

	defaultValue = property(lambda self: self.get("/DV"))
	"""
	Read-only property accessing the default value of this field.
	"""

	additionalActions = property(lambda self: self.get("/AA"))
	"""
	Read-only property accessing the additional actions dictionary.
	This dictionary defines the field's behavior in response to trigger events.
	See Section 8.5.2 of the PDF 1.7 reference.
	"""


class Destination(TreeObject):
	"""
	A class representing a destination within a PDF file.
	See section 8.2.1 of the PDF 1.6 reference.

	:param str title: Title of this destination.
	:param int page: Page number of this destination.
	:param str typ: How the destination is displayed.
	:param args: Additional arguments may be necessary depending on the type.
	:raises PdfReadError: If destination type is invalid.

	Valid ``typ`` arguments (see PDF spec for details):
			 /Fit		No additional arguments
			 /XYZ		[left] [top] [zoomFactor]
			 /FitH		[top]
			 /FitV		[left]
			 /FitR		[left] [bottom] [right] [top]
			 /FitB		No additional arguments
			 /FitBH		[top]
			 /FitBV		[left]
	"""
	def __init__(self, title, page, typ, *args):
		DictionaryObject.__init__(self)
		self[NameObject("/Title")] = title
		self[NameObject("/Page")] = page
		self[NameObject("/Type")] = typ

		# from table 8.2 of the PDF 1.7 reference.
		if typ == "/XYZ":
			(self[NameObject("/Left")], self[NameObject("/Top")],
				self[NameObject("/Zoom")]) = args
		elif typ == "/FitR":
			(self[NameObject("/Left")], self[NameObject("/Bottom")],
				self[NameObject("/Right")], self[NameObject("/Top")]) = args
		elif typ in ["/FitH", "/FitBH"]:
			self[NameObject("/Top")], = args
		elif typ in ["/FitV", "/FitBV"]:
			self[NameObject("/Left")], = args
		elif typ in ["/Fit", "/FitB"]:
			pass
		else:
			raise utils.PdfReadError("Unknown Destination Type: %r" % typ)

	def getDestArray(self):
		return ArrayObject([self.raw_get('/Page'), self['/Type']] + [self[x] for x in ['/Left', '/Bottom', '/Right', '/Top', '/Zoom'] if x in self])

	def writeToStream(self, stream, encryption_key):
		stream.write(b_("<<\n"))
		key = NameObject('/D')
		key.writeToStream(stream, encryption_key)
		stream.write(b_(" "))
		value = self.getDestArray()
		value.writeToStream(stream, encryption_key)

		key = NameObject("/S")
		key.writeToStream(stream, encryption_key)
		stream.write(b_(" "))
		value = NameObject("/GoTo")
		value.writeToStream(stream, encryption_key)

		stream.write(b_("\n"))
		stream.write(b_(">>"))

	title = property(lambda self: self.get("/Title"))
	"""
	Read-only property accessing the destination title.

	:rtype: str
	"""

	page = property(lambda self: self.get("/Page"))
	"""
	Read-only property accessing the destination page number.

	:rtype: int
	"""

	typ = property(lambda self: self.get("/Type"))
	"""
	Read-only property accessing the destination type.

	:rtype: str
	"""

	zoom = property(lambda self: self.get("/Zoom", None))
	"""
	Read-only property accessing the zoom factor.

	:rtype: int, or ``None`` if not available.
	"""

	left = property(lambda self: self.get("/Left", None))
	"""
	Read-only property accessing the left horizontal coordinate.

	:rtype: int, or ``None`` if not available.
	"""

	right = property(lambda self: self.get("/Right", None))
	"""
	Read-only property accessing the right horizontal coordinate.

	:rtype: int, or ``None`` if not available.
	"""

	top = property(lambda self: self.get("/Top", None))
	"""
	Read-only property accessing the top vertical coordinate.

	:rtype: int, or ``None`` if not available.
	"""

	bottom = property(lambda self: self.get("/Bottom", None))
	"""
	Read-only property accessing the bottom vertical coordinate.

	:rtype: int, or ``None`` if not available.
	"""


class Bookmark(Destination):
	def writeToStream(self, stream, encryption_key):
		stream.write(b_("<<\n"))
		for key in [NameObject(x) for x in ['/Title', '/Parent', '/First', '/Last', '/Next', '/Prev'] if x in self]:
			key.writeToStream(stream, encryption_key)
			stream.write(b_(" "))
			value = self.raw_get(key)
			value.writeToStream(stream, encryption_key)
			stream.write(b_("\n"))
		key = NameObject('/Dest')
		key.writeToStream(stream, encryption_key)
		stream.write(b_(" "))
		value = self.getDestArray()
		value.writeToStream(stream, encryption_key)
		stream.write(b_("\n"))
		stream.write(b_(">>"))


def encode_pdfdocencoding(unicode_string):
	retval = b_('')
	for c in unicode_string:
		try:
			retval += b_(chr(_pdfDocEncoding_rev[c]))
		except KeyError:
			raise UnicodeEncodeError("pdfdocencoding", c, -1, -1,
					"does not exist in translation table")
	return retval


def decode_pdfdocencoding(byte_array):
	retval = u_('')
	for b in byte_array:
		c = _pdfDocEncoding[ord_(b)]
		if c == u_('\u0000'):
			raise UnicodeDecodeError("pdfdocencoding", utils.barray(b), -1, -1,
					"does not exist in translation table")
		retval += c
	return retval

_pdfDocEncoding = (
  u_('\u0000'), u_('\u0000'), u_('\u0000'), u_('\u0000'), u_('\u0000'), u_('\u0000'), u_('\u0000'), u_('\u0000'),
  u_('\u0000'), u_('\u0000'), u_('\u0000'), u_('\u0000'), u_('\u0000'), u_('\u0000'), u_('\u0000'), u_('\u0000'),
  u_('\u0000'), u_('\u0000'), u_('\u0000'), u_('\u0000'), u_('\u0000'), u_('\u0000'), u_('\u0000'), u_('\u0000'),
  u_('\u02d8'), u_('\u02c7'), u_('\u02c6'), u_('\u02d9'), u_('\u02dd'), u_('\u02db'), u_('\u02da'), u_('\u02dc'),
  u_('\u0020'), u_('\u0021'), u_('\u0022'), u_('\u0023'), u_('\u0024'), u_('\u0025'), u_('\u0026'), u_('\u0027'),
  u_('\u0028'), u_('\u0029'), u_('\u002a'), u_('\u002b'), u_('\u002c'), u_('\u002d'), u_('\u002e'), u_('\u002f'),
  u_('\u0030'), u_('\u0031'), u_('\u0032'), u_('\u0033'), u_('\u0034'), u_('\u0035'), u_('\u0036'), u_('\u0037'),
  u_('\u0038'), u_('\u0039'), u_('\u003a'), u_('\u003b'), u_('\u003c'), u_('\u003d'), u_('\u003e'), u_('\u003f'),
  u_('\u0040'), u_('\u0041'), u_('\u0042'), u_('\u0043'), u_('\u0044'), u_('\u0045'), u_('\u0046'), u_('\u0047'),
  u_('\u0048'), u_('\u0049'), u_('\u004a'), u_('\u004b'), u_('\u004c'), u_('\u004d'), u_('\u004e'), u_('\u004f'),
  u_('\u0050'), u_('\u0051'), u_('\u0052'), u_('\u0053'), u_('\u0054'), u_('\u0055'), u_('\u0056'), u_('\u0057'),
  u_('\u0058'), u_('\u0059'), u_('\u005a'), u_('\u005b'), u_('\u005c'), u_('\u005d'), u_('\u005e'), u_('\u005f'),
  u_('\u0060'), u_('\u0061'), u_('\u0062'), u_('\u0063'), u_('\u0064'), u_('\u0065'), u_('\u0066'), u_('\u0067'),
  u_('\u0068'), u_('\u0069'), u_('\u006a'), u_('\u006b'), u_('\u006c'), u_('\u006d'), u_('\u006e'), u_('\u006f'),
  u_('\u0070'), u_('\u0071'), u_('\u0072'), u_('\u0073'), u_('\u0074'), u_('\u0075'), u_('\u0076'), u_('\u0077'),
  u_('\u0078'), u_('\u0079'), u_('\u007a'), u_('\u007b'), u_('\u007c'), u_('\u007d'), u_('\u007e'), u_('\u0000'),
  u_('\u2022'), u_('\u2020'), u_('\u2021'), u_('\u2026'), u_('\u2014'), u_('\u2013'), u_('\u0192'), u_('\u2044'),
  u_('\u2039'), u_('\u203a'), u_('\u2212'), u_('\u2030'), u_('\u201e'), u_('\u201c'), u_('\u201d'), u_('\u2018'),
  u_('\u2019'), u_('\u201a'), u_('\u2122'), u_('\ufb01'), u_('\ufb02'), u_('\u0141'), u_('\u0152'), u_('\u0160'),
  u_('\u0178'), u_('\u017d'), u_('\u0131'), u_('\u0142'), u_('\u0153'), u_('\u0161'), u_('\u017e'), u_('\u0000'),
  u_('\u20ac'), u_('\u00a1'), u_('\u00a2'), u_('\u00a3'), u_('\u00a4'), u_('\u00a5'), u_('\u00a6'), u_('\u00a7'),
  u_('\u00a8'), u_('\u00a9'), u_('\u00aa'), u_('\u00ab'), u_('\u00ac'), u_('\u0000'), u_('\u00ae'), u_('\u00af'),
  u_('\u00b0'), u_('\u00b1'), u_('\u00b2'), u_('\u00b3'), u_('\u00b4'), u_('\u00b5'), u_('\u00b6'), u_('\u00b7'),
  u_('\u00b8'), u_('\u00b9'), u_('\u00ba'), u_('\u00bb'), u_('\u00bc'), u_('\u00bd'), u_('\u00be'), u_('\u00bf'),
  u_('\u00c0'), u_('\u00c1'), u_('\u00c2'), u_('\u00c3'), u_('\u00c4'), u_('\u00c5'), u_('\u00c6'), u_('\u00c7'),
  u_('\u00c8'), u_('\u00c9'), u_('\u00ca'), u_('\u00cb'), u_('\u00cc'), u_('\u00cd'), u_('\u00ce'), u_('\u00cf'),
  u_('\u00d0'), u_('\u00d1'), u_('\u00d2'), u_('\u00d3'), u_('\u00d4'), u_('\u00d5'), u_('\u00d6'), u_('\u00d7'),
  u_('\u00d8'), u_('\u00d9'), u_('\u00da'), u_('\u00db'), u_('\u00dc'), u_('\u00dd'), u_('\u00de'), u_('\u00df'),
  u_('\u00e0'), u_('\u00e1'), u_('\u00e2'), u_('\u00e3'), u_('\u00e4'), u_('\u00e5'), u_('\u00e6'), u_('\u00e7'),
  u_('\u00e8'), u_('\u00e9'), u_('\u00ea'), u_('\u00eb'), u_('\u00ec'), u_('\u00ed'), u_('\u00ee'), u_('\u00ef'),
  u_('\u00f0'), u_('\u00f1'), u_('\u00f2'), u_('\u00f3'), u_('\u00f4'), u_('\u00f5'), u_('\u00f6'), u_('\u00f7'),
  u_('\u00f8'), u_('\u00f9'), u_('\u00fa'), u_('\u00fb'), u_('\u00fc'), u_('\u00fd'), u_('\u00fe'), u_('\u00ff')
)

assert len(_pdfDocEncoding) == 256

_pdfDocEncoding_rev = {}
for i in range(256):
	char = _pdfDocEncoding[i]
	if char == u_("\u0000"):
		continue
	assert char not in _pdfDocEncoding_rev
	_pdfDocEncoding_rev[char] = i



#copyright ReportLab Europe Limited. 2000-2016
"Utilities for sending mail"

def _sendMail(smtpServer,sender,receiver,msg, user=None, password=None):
    import smtplib
    server = smtplib.SMTP(smtpServer)
    try:
        if user:
            server.login(user, password)
        server.sendmail(sender, receiver, msg)
    finally:
        server.quit()

import re
few_re = re.compile( '^=\\?.*\\?=$' )

def _encmin(u):
    """minimally encode unicode"""
    for enc in ('utf-8','iso-8859-1'):
        try:
            return u.encode(enc), enc
        except UnicodeEncodeError:
            pass
    return u.encode('utf-8'), 'utf-8'

def _xencmin(u, force=False):
    '''
    Used for minimally encoding a unicode into whatever is shortest.
    '''
    high_count = len([ x for x in u if ord(x) > 127])
    if not force:
        if high_count == 0:
            return u,'7bit'
    base64_count = (len(u)*4)/3
    if base64_count % 4:
        base64_count += 4 - ( base64_count % 4 )
    qp_count = len(u) + high_count * 2
    if base64_count < qp_count:
        return u.encode('base64'),'Base64'
    enc = '*'
    if '*' in u:
        if '^' not in u:
            enc = '^'
        elif '~' not in u:
            enc = '~'
        else:
            enc = None
    if enc:
        u = u.replace(' ',enc)
        u = u.encode('quoted-printable')
        u = u.replace(enc,' ')
        return u, 'Quoted-Printable'
    return u.encode('quoted-printable').replace('=20',' '),'Quoted-Printable'

def _encoded_word(word, enc='utf8', force=False):
    if not isinstance(word,str):
        word = word.decode(enc)
    wchr, charset = _encmin(word)
    force = force or few_re.match(wchr)
    wenc, enc = _xencmin(wchr, force=force)
    if enc == '7bit':
        return wenc,False
    else:
        menc = 'X'
        if enc == 'Base64':
            menc = 'B'
            wenc = ''.join(wenc.split('\n'))
        elif enc == 'Quoted-Printable':
            menc = 'Q'
            wenc = ''.join(wenc.split( '=\n' ))
            wenc = wenc.replace(' ','_')
            wenc = wenc.replace('?','=%X'%ord( '?' ) )
        return '=?%s?%s?%s?=' % (charset.upper(),menc,wenc), True

def encoded_word(word,enc='utf8',force=False):
    '''attempt a good encoded-word encoding of word http://en.wikipedia.org/wiki/MIME#Encoded-Word'''
    return _encoded_word(word,enc,force)[0]

def sendMail(smtpServer, sender, receiver, subject, text, fromaddr=None, user=None, password=None):
    "Sends a simple mail"
    if fromaddr is None: fromaddr=sender
    _sendMail(smtpServer,sender,receiver,
            "To: %s\nFrom: %s\nSubject: %s\n\n%s" % (receiver,fromaddr,subject, text), user=user, password=password)

class MailAttachment:
    def __init__(self,name,stream=None, mimeType=('application','octet-stream')):
        self.stream = stream
        self.name = name
        self.mimeType = mimeType
    def read(self):
        if self.stream:
            return self.stream.read()
        else:
            return open(self.name,'rb').read()

def sendEmail(smtpServer, toAddress, fromAddress, subject, msgBody, subjectEncoding=None, sender=None, user=None, password=None, attachments=[], isHtmlText=False):
    import os
    from email.MIMEText import MIMEText
    text = MIMEText(msgBody, (isHtmlText and 'html' or 'plain'), 'utf-8') # Assume text is UTF8
    if not attachments:
        msg = text
    else:
        from email.MIMEMultipart import MIMEMultipart
        from email.MIMEBase import MIMEBase
        from email import Encoders
        msg = MIMEMultipart()
        msg.attach(text)
        for f in attachments:
            name = os.path.basename(f.name)
            part = MIMEBase(*f.mimeType)
            part.set_payload(f.read())
            Encoders.encode_base64(part)
            part.add_header('Content-Disposition', 'attachment; filename="%s"' % name )
            msg.attach(part)
    if subjectEncoding is not None:
        from email.Header import Header
        subject = Header(subject,subjectEncoding)
    msg['Subject'] = subject
    msg['From'] = fromAddress
    msg['To'] = toAddress
    if sender: msg['Sender'] = sender
    _sendMail(smtpServer,fromAddress,(toAddress,), msg.as_string(), user=user, password=password)

def sendHtmlMail(smtpServer, sender, receiver, subject, html, user=None, password=None):
    "Sends an html email (html argument is an HTML string)"
    import MimeWriter
    import mimetools
    import io

    # Produce an approximate textual rendering of the HTML string,
    # unless you have been given a better version as an argument
    import htmllib, formatter
    textout = io.StringIO()
    formtext = formatter.AbstractFormatter(formatter.DumbWriter(textout))
    parser = htmllib.HTMLParser(formtext)
    parser.feed(html)
    parser.close()
    text = textout.getvalue()
    del textout, formtext, parser

    out = io.StringIO() # output buffer for our message
    htmlin = io.StringIO(html)
    txtin = io.StringIO(text)

    writer = MimeWriter.MimeWriter(out)

    # Set up some basic headers. Place subject here
    # because smtplib.sendmail expects it to be in the
    # message body, as relevant RFCs prescribe.
    writer.addheader("Subject", subject)
    writer.addheader("MIME-Version", "1.0")

    # Start the multipart section of the message.
    # Multipart/alternative seems to work better
    # on some MUAs than multipart/mixed.
    writer.startmultipartbody("alternative")
    writer.flushheaders()

    # the plain-text section: just copied through, assuming iso-8859-1
    subpart = writer.nextpart()
    pout = subpart.startbody("text/plain", [("charset", 'iso-8859-1')])
    pout.write(txtin.read())
    txtin.close()

    # the HTML subpart of the message: quoted-printable, just in case
    subpart = writer.nextpart()
    subpart.addheader("Content-Transfer-Encoding", "quoted-printable")
    pout = subpart.startbody("text/html", [("charset", 'us-ascii')])
    mimetools.encode(htmlin, pout, 'quoted-printable')
    htmlin.close()

    # You're done; close your writer and return the message body
    writer.lastpart()
    msg = out.getvalue()
    out.close()
    smtpServer._sendMail(smtpServer,sender, receiver, msg, user=user, password=password)

# any other utilities
def validateEmail(addr):
    "regex that checks for email validity( '@' sign etc..)"
    import re
    email = re.compile(r"^[a-z0-9\._-]+@+[a-z0-9\._-]+\.+[a-z]{2,3}$",re.I)
    return email.match(addr)

def escapeHTML(s):
    """
    Escape all characters with a special meaning in HTML
    to appropriate character tags
    """
    _ESCAPE_HTML_CHARS=list('&<>":={}()')
    _ESCAPE_HTML_CHARS_TRANS = [(c,'&#%d;' % ord(c)) for c in _ESCAPE_HTML_CHARS]    
    for c,e in _ESCAPE_HTML_CHARS_TRANS:
        s = s.replace(c,e)
    return s



"""XML Data Acquisition Languagelet.

Basic data acquisition tasks which could be done with Python
scriptlets are provided as XML tags.

"""
try:
    import urllib.request
    urllib_urlopen = urllib.request.urlopen
except ImportError:
    import urllib
    urllib_urlopen = urllib.urlopen

from rlextra.radxml.xmlutils import TagWrapper
from reportlab.lib.rparsexml import parsexml

class DataFetcher:
    def __init__(self, **kw):
        self.name = None
        self.assignTo = None
        self.children = []
        for key, value in list(kw.items()):
            setattr(self, key, value)

    def getData(self):
        "Override this to retrieve the data"
        return None

    def fetch(self, nameSpace):
        "This is called by external environment"
        data = self.getData()
        if self.assignTo:
            parent = nameSpace[self.assignTo]
            setattr(parent, self.name, data)
        else:
            nameSpace[self.name] = data


class XmlDataFetcher(DataFetcher):
    "Read xml file from a file or url"
    #needs support for dtdDir etc.
    def getData(self):
        if hasattr(self, 'url'):
            xmlText = urllib_urlopen(self.url).read()
        elif hasattr(self, 'fileName'):
            xmlText = open(self.fileName,'rb').read()
        elif type(self.children[0]) is type(''):
            xmlText = self.children[0]
        else:
            raise ValueError("xml tag must have a 'fileName' or 'url' attribute")
        tree = parsexml(xmlText)
        wrapped = TagWrapper(tree)
        return wrapped

tagMap = {'xmlData':XmlDataFetcher}

def acquireData(nodeList, nameSpace):
    """Convert nodelist to relevant fetchers and get data"""
    for elem in nodeList:
        if isinstance(elem,tuple):
            (tagName, attrs, children, stuff) = elem
            klass = tagMap[tagName]
            obj = klass(**attrs)
            obj.children = children
            obj.fetch(nameSpace)

        



__code__='Gas1]gMYJ*&Daab9@M6,Tl2\'3RkPIM.8URB#Y5%a8jc$n5q^6!^"_,fZ/+51Rp,2XqT=T_97GtUG[eZeBXl(8U[cXRR@VkJIj8I<=]d=baUiu.EQB*9-mNAXOb)Xk:\'\\_2rE!66)Uk+j$EV<@XNW"nKh\\j&]>UT!1n.p[?`$\'R"0tTaeNR`.i=pO!;f;*A5c<[NH\'<%^>6Qq*/tW`R`+g<%c"D.I+$YWH3m\'PbLU)ju)O,olB@P<HKqs<ni,&;e0ZP$5#3ILG$hT+tP>bbL\'69g\'D4o\'@7<;\\RUF>Lt9C1G,#MO]+BG>]NFunmEP4%W4\\*^Tco\'8/+cUtfSGtNpp,_ni4&OE3D,;USUG;:(JH<9Mqg33+RLb_s0DMaJLQ143)YTMPVO6]^=p#/rbVh.arHeWP]m,U.TOp&CYQq(hX<oHW09fkotg4gMWf8[EqGMT@[]%rbH6F<erHdMB$F-!7bC0l,n-4o[41h9*)dr`QVc]_J0QFC_6[amNP+AL/j".r)H1G]WYmjg3N)2m6J.X>4$*uuBT,2WgBD;1,`[[Gjo6SIeO.Yo<*(+/JtJH-TQ:6\'hc^fh;!V\\\'<a9@Tp"^E_!#j?Fl,r78t5(D>&fin+qbkS45++$8DBkmsN-7N<if]"[[hq4h(#Gg3$3r&mhrgR25M\\A&%DV>H;p/j4HBc<]MFN>*1k(T+Zo0(*#l@_L?sEDN#j1S+(1\\A\'+R];&S%k=4&Z]$OY:Xo5mfN&Gmm+QP)UN-=7[FKWO[Sd<iN\\IlbghQo*,kV#"<J9K*j!7%j3isOd<$_=<u.rBr^fPH"MJD!1`WsS6s`bREg+kO_0WT%S+8!:Ijd(cI!B8Xhr7<oK1M\\!b&Jb%!Rk5rP`cDM<<RK^EWksaP<*)MQ*[/\\kdB14a\\gJoe7jFsX6bb4eaRPS+.XQ6s!(]m^_C`mXs#ScaMN:(?N:nY\\rUq`lS.mH,IeT4ENGO\\m(Vi!oX;!gJC9q.i2p&d,#>bV~>'



#pageflow.py
"""Support for using PageCatcher within Platypus"""
__all__ = ('LoadPdfFlowable', 'ShowPdfFlowable', 'loadPdf',
            )
import os
from reportlab.lib.utils import pickle
from reportlab.platypus.flowables import Flowable, PageBreak
from rlextra.pageCatcher.pageCatcher import storeFormsInMemory,\
        restoreFormsInMemory, restoreFormsFromDict, fileName2Prefix
from reportlab.pdfbase.pdfdoc import xObjectName
from reportlab.platypus.doctemplate import NextPageTemplate, NotAtTopPageBreak
from reportlab.lib.utils import open_and_read, asUnicodeEx, annotateException
from reportlab.graphics.shapes import transformPoints

class LoadPdfFlowable(Flowable):
    """Imports PDF content into the current canvas when drawn.

    Used by RML2PDF, which has to store the forms itself but defer
    placing them into a document"""
    def __init__(self, pickledStuff, isDict=False):
        Flowable.__init__(self)
        self.pickledStuff = pickledStuff
        self._isDict = isDict

    def drawOn(self, canvas, x, y, _sW=0):
        "This makes sure the form XObjects are loaded into the canvas"
        (self._isDict and restoreFormsFromDict or restoreFormsInMemory)(self.pickledStuff,
                             canvas,
                             allowDuplicates=1)

def _formName2ExtraInfo(canv,name):
    doc = canv._doc
    internalname = xObjectName(name)
    if internalname in doc.idToObject:
        form = doc.idToObject[internalname]
        if hasattr(form,'_extra_pageCatcher_info'):
            return form._extra_pageCatcher_info

def normalizeBoxName(boxName,_normalizedBoxNames=dict(
                                mediabox='MediaBox',
                                cropbox='CropBox',
                                artbox='ArtBox',
                                trimbox='TrimBox',
                                bleedbox='BleedBox',
                                )):
    if boxName is None: return 'MediaBox'
    return _normalizedBoxNames[boxName.lower()]

def _formName2OrientationMatrix(canv,name,orientation,pdfBoxType,pageSize,xi,pageSizeHandler):
    if xi:
        cps = canv._pagesize
        po = xi.get('Rotate',0)
        if isinstance(pdfBoxType,(list,tuple)):
            x0,y0,x1,y1 = pdfBoxType
        else:
            x0,y0,x1,y1 = xi.get(normalizeBoxName(pdfBoxType),xi['MediaBox'])
        xscale = yscale = 1
        w = abs(x1-x0)
        h = abs(y1-y0)
        if pageSize=='set' or isinstance(pageSize,(list,tuple)):
            if pageSize=='set':
                nps = (w,h)
            else:
                nps = abs(pageSize[2]-pageSize[0]),abs(pageSize[3]-pageSize[1])
            canv.setPageSize(nps)
            if pageSizeHandler:
                if pageSizeHandler.first:
                    pageSizeHandler.oldPageSize = cps
                else:
                    canv._hanging_pagesize = pageSizeHandler.oldPageSize
        elif pageSize in ('fit','orthofit'):
            xscale = cps[0]/w
            yscale = cps[1]/h
            if pageSize=='orthofit':
                xscale = yscale = min(xscale,yscale)
            x0 *= xscale
            y0 *= yscale
            x1 *= xscale
            y1 *= yscale
        elif pageSize in ('center','centre'):
            dx = (cps[0] - w) * 0.5
            dy = (cps[1] - h) * 0.5
            x0 -= dx
            x1 -= dx
            y0 -= dy
            y1 -= dy
        if orientation=='auto':
            orientation=po
        elif orientation is None:
            orientation = 0
        orientation = int(orientation % 360)
        if orientation==0:
            matrix = [xscale,0,0,yscale,-x0,-y0]
        elif orientation==90:
            matrix = [0,-yscale,xscale,0,-y0,x1-x0]
        elif orientation==180:
            matrix = [-xscale,0,0,-yscale,x1-x0,y1-y0]
        else:
            matrix = [0,yscale,-xscale,0,y1-y0,-x0]
        return matrix if matrix != [1,0,0,1,0,0] else None

class ShowPdfFlowable(Flowable):
    "draws a form xobject in absolute coordinates"
    def __init__(self, formName, orientation=None,iptrans=None, callback=None,
            pdf_data=None, user_data=None,
            pdfBoxType=None,
            autoCrop=None,
            pageSize=None,
            pageSizeHandler=None,
            ):
        Flowable.__init__(self)
        self.formName = formName
        self._orientation = orientation
        self._iptrans = iptrans
        self._pdfBoxType = pdfBoxType
        self._autoCrop = autoCrop
        self._pageSize = pageSize
        self._pageSizeHandler = pageSizeHandler
        self._pdf_data = pdf_data
        if callback:
            self.callback = lambda key,canv: callback(key,canv,self,pdf_data,user_data)
        else:
            self.callback = lambda key,canv: None

    def drawOn(self, canvas, x, y, _sW=0):
        iptrans = self._iptrans
        matrix = self._orientation
        pdfBoxType = self._pdfBoxType
        autoCrop = self._autoCrop
        pageSize = self._pageSize
        doTrans = iptrans and not iptrans.trivial()
        self.callback('raw-pre',canvas)
        restore = doTrans or autoCrop or pdfBoxType or matrix!=None or pageSize
        if restore:
            xi = _formName2ExtraInfo(canvas,self.formName)
            canvas.saveState()
            if pdfBoxType or matrix is not None or pageSize:
                matrix = _formName2OrientationMatrix(canvas,self.formName,matrix,pdfBoxType,pageSize,xi,self._pageSizeHandler)
                if matrix:
                    canvas.transform(*matrix)
            if iptrans:
                if not iptrans.noRotate():
                    canvas.rotate(iptrans.degrees)
                if not iptrans.noTranslate():
                    canvas.translate(iptrans.dx,iptrans.dy)
                if not iptrans.noScale():
                    canvas.scale(iptrans.sx,iptrans.sy)
            self.callback('transformed-pre',canvas)

        if autoCrop:
            if isinstance(autoCrop,(list,tuple)):
                clip = list(map(float,autoCrop))
            elif isinstance(autoCrop,str):
                clip = xi[normalizeBoxName(autoCrop)]
            else:
                clip = xi['CropBox']
            P = (clip[0],clip[1]),(clip[0],clip[3]),(clip[2],clip[3]),(clip[2],clip[1])
            #if canvas._code and canvas._code[-1][-3:]==' cm':
            #   L = canvas._code[-1].split()
            #   transform = list(map(float,L[-7:-1]))
            #   P = transformPoints(transform,P)
            path = canvas.beginPath()
            path.moveTo(*P[0])
            path.lineTo(*P[1])
            path.lineTo(*P[2])
            path.lineTo(*P[3])
            path.close()
            canvas.clipPath(path,0,0)

        try:
            #this won't do much if already loaded
            canvas.doForm(self.formName)
        except:
            annotateException('\ncnvas.doForm(%r) error handling %r' % (self.formName,self._pdf_data))

        if restore:
            self.callback('transformed-post',canvas)
            canvas.restoreState()
        self.callback('raw-post',canvas)

class IPTrans:
    def __init__(self,sx,sy,dx,dy,degrees):
        self.sx = sx
        self.sy = sy
        self.dx = dx
        self.dy = dy
        self.degrees = degrees

    def trivial(self):
        return self.noScale() and self.noTranslate() and self.noRotate()

    def noScale(self):
        return not (abs(self.sx-1)>1e-8 or abs(self.sy-1)>1e-8)

    def noTranslate(self):
        return not (abs(self.dx)>1e-8 or abs(self.dy)>1e-8)

    def noRotate(self):
        return not abs(self.degrees)>1e-8

def expandPageNumbers(txt):
    """Convert expression like '1,3,7-10,12' to a list of numbers"""
    if not txt: return None

    chunks = txt.split(',')
    pages = []
    pages_append = pages.append
    for chunk in chunks:
        chunk = chunk.strip()
        bits = chunk.split('-')
        if len(bits) == 1:
            pages_append(int(chunk))
        elif len(bits) == 2:
            start = int(bits[0])
            end = int(bits[1])
            for pg in range(start, end+1):
                pages_append(pg)
        else:
            pass
    return pages

class OutlineEntry(Flowable):
    """make an outline entry (lazily convert content)
       should work either in story or in page graphic mode!
    """
    _ZEROSIZE = 1
    _N = 0
    def __init__(self, level, content, closed, name=None, newBookmark=1):
        self.level = level
        self.content = content
        self.closed = closed
        if name is None:
            name = "k%r%d" % (id(self),self._N)     # arbitrary unique string
            self._N += 1
            newBookmark = 1
        self.name = name
        self.newBookmark = newBookmark

    def wrap(self, w, h):
        return 0,0 # consumes no space
    def draw(self):
        "flowable mode"
        self.doEntry(self.canv)
    def __call__(self, canv, doc):
        "graphic mode"
        self.doEntry(canv)
    def doEntry(self, canv):
        content = "".join(map(asUnicodeEx, self.content))
        name = self.name
        if self.newBookmark:
            canv.bookmarkPage(name)
        # bogus level may result in error here
        canv.addOutlineEntry(content, name, self.level, self.closed)
    def __str__(self):
        return ''

def includePdfFlowables(fileName,
                        pages=None,
                        dx=0, dy=0, sx=1, sy=1, degrees=0,
                        orientation=None,
                        isdata=False,       #True if this is a preprocessed data file
                        leadingBreak=True,  #True/False or 'notattop'
                        template=None,
                        outlineText=None,
                        outlineLevel=0,
                        outlineClosed=0,
                        pdfBoxType = None,
                        autoCrop = False,
                        pageSize=None,
                        callback=None,
                        user_data=None,
                        ):
    '''
    includePdfFlowables creates a list of story flowables that
                        represents an included PDF.
    Arguments       meaning
    fileName        string name of a .pdf or .data file
    pages           If None all pages will be used, else this argument can
                    be a string like '1,2,4-6,12-10,15' or an explicit
                    list of integers eg [1,2,7].

    dx,dy,          translation together all these make up a transformation
    sx,sy,          scaling     matrix
    degrees,        rotation

    orientation     None or integer degrees eg 0 90 270 or 'portrait'/'landscape'
    isdata          True if fileName argument refers to a .data file (as
                    produced by pageCatcher)
    leadingBreak    True/False or 'notattop' specifies whether a leading
                    page break should be used; 'notattop' means a page break
                    will not be used if the story is at the top of a frame.
    template        If specified the index or name of a template to be used.
    outlineText     Any outline text to be used (default None)
    outlineLevel    The level of any outline text.
    outlineClosed   True/False if the outline should be closed or open.

    pdfBoxType      which box to use or None or [x0,y0,  x1,y1]

    autoCrop        True/False crop/don't crop with CropBox (default is False)
                    boxname use for cropping
                    [x0,y0,  x1,y1] crop area

    pageSize        default None ie leave page size alone
                    'set' adjust page size to incoming box
                    'fit' scale incoming box to fit page size
                    'orthfit' orthogonally scale incoming box to fit
                    'center' or 'centre' center the incoming box in
                    the existing page size
                    [x0,y0, x1,y1] use this as the page size

    callback        draw time callback with signature

                    callback(canvas,key,obj,pdf_data,user_data)

                    canvas the canvas being drawn on
                    key may be 'raw-pre'|'transformed-pre'|'transformed-post'|'raw-post'
                    obj the flowable calling the callback
                    pdf_data ('fileName',pageNumber)
                    user_data user data passed down to the flowable from
                              IncludePdfFlowable.

    user_data       information to be passed to the callback
    '''
    try:
        orientation=int(orientation)
        orientation = orientation % 360
    except:
        if orientation=='portrait':
            orientation = 0
        elif orientation=='landscape':
            orientation = 90
        elif orientation!='auto' and orientation!=None:
            raise ValueError('Bad value %r for orientation attribute' % orientation)

    iptrans = IPTrans(sx,sy,dx,dy,degrees)
    if iptrans.trivial(): iptrans = None

    pages = expandPageNumbers(pages)

    # this one is unusual in that it returns a list of objects to
    # go into the story.
    output = []
    output_append = output.append

    if template:
        output_append(NextPageTemplate(template))

    try:
        if isdata:
            pickledStuff = pickle.loads(open_and_read(fileName))
            formNames = pickledStuff[None]
        else:
            #read in the PDF file right now and get the pickled object
            # and names
            pdfContent = open_and_read(fileName)
            prefix = fileName2Prefix(fileName)
            (formNames, pickledStuff) = storeFormsInMemory(
                    pdfContent,
                    prefix=prefix,
                    all=1,
                    BBoxes=0,
                    extractText=0,
                    fformname=None)
    except:
        annotateException('\nerror storing %r in memory\n' % fileName)

    #if explicit pages requested, slim it down.
    if pages:
        newNames = []
        for pgNo in pages:
            newNames.append(formNames[pgNo-1])
        formNames = newNames

    #make object 1 for story
    loader = LoadPdfFlowable(pickledStuff,isdata)
    output_append(loader)

    #now do first page.  This is special as it might
    #have an outline
    formName = formNames[0]
    if leadingBreak:
        output_append((leadingBreak=='notattop' and NotAtTopPageBreak or PageBreak)())
    if outlineText:
        output_append(OutlineEntry(outlineLevel, outlineText, outlineClosed))

    if pageSize=='fit':
        class PageSizeHandler(object):
            '''simple class to allow communications between first and last ShowPdfFlowables'''
            _oldPageSize = [None]
            def __init__(self,first):
                self.first = first

            def oldPageSize(self,v):
                self._oldPageSize[0] = v
            oldPageSize = property(lambda self: self._oldPageSize[0],oldPageSize)
        pageSizeHandler = PageSizeHandler(True)
    else:
        pageSizeHandler = None
    output_append(ShowPdfFlowable(formName,orientation=orientation,iptrans=iptrans,
                        callback=callback,
                        pdf_data=(fileName,pages[0] if pages else 1),
                        user_data=user_data,
                        pdfBoxType=pdfBoxType,
                        autoCrop=autoCrop,
                        pageSize=pageSize,
                        pageSizeHandler=pageSizeHandler,
                        ))

    #now make a shower for each laterpage, and a page break
    for i,formName in enumerate(formNames[1:]):
        i += 1
        output_append(PageBreak())
        output_append(ShowPdfFlowable(formName,orientation=orientation,iptrans=iptrans,
                callback=callback,
                pdf_data=(fileName,pages[i] if pages else i),
                user_data=user_data,
                pdfBoxType=pdfBoxType,
                autoCrop=autoCrop,
                pageSize=pageSize,
                pageSizeHandler=None,
                ))
    if pageSize=='fit':
        output[-1]._pageSizeHandler = PageSizeHandler(False)
    return output

class IncludePdfFlowable(Flowable):
    '''
    includePdfFlowable flowabnle that splits to a list of story flowables that
                        represents an included PDF.
    Arguments       meaning
    fileName        string name of a .pdf or .data file
    pages           If None all pages will be used, else this argument can
                    be a string like '1,2,4-6,12-10,15' or an explicit
                    list of integers eg [1,2,7].

    dx,dy,          translation together all these make up a transformation
    sx,sy,          scaling     matrix
    degrees,        rotation

    orientation     None or integer degrees eg 0 90 270 or 'portrait'/'landscape'
    isdata          True if fileName argument refers to a .data file (as
                    produced by pageCatcher)
    leadingBreak    True/False or 'notattop' specifies whether a leading
                    page break should be used; 'notattop' means a page break
                    will not be used if the story is at the top of a frame.
    template        If specified the index or name of a template to be used.
    outlineText     Any outline text to be used (default None)
    outlineLevel    The level of any outline text.
    outlineClosed   True/False if the outline should be closed or open.

    pdfBoxType      which box to use or None or [x0,y0,  x1,y1]

    autoCrop        True/False crop/don't crop with CropBox (default is False)
                    boxname use for cropping
                    [x0,y0,  x1,y1] crop area

    pageSize        default None ie leave page size alone
                    'set' adjust page size to incoming box
                    'fit' scale incoming box to fit page size
                    'orthfit' orthogonally scale incoming box to fit
                    'center' or 'centre' center the incoming box in
                    the existing page size
                    [x0,y0, x1,y1] use this as the page size

    callback        draw time callback with signature

                    callback(canvas,key,obj,pdf_data,user_data)

                    canvas the canvas being drawn on
                    key may be 'raw-pre'|'transformed-pre'|'transformed-post'|'raw-post'
                    obj the flowable calling the callback
                    pdf_data ('fileName',pageNumber)
                    user_data user data passed down to the flowable from
                              IncludePdfFlowable.

    user_data       information to be passed to the callback
    '''
    def __init__(self,
                fileName,
                pages=None,
                dx=0, dy=0, sx=1, sy=1, degrees=0,
                orientation=None,
                isdata=False,       #True if this is a preprocessed data file
                leadingBreak=True,  #True/False or 'notattop'
                template=None,
                outlineText=None,
                outlineLevel=0,
                outlineClosed=0,
                pdfBoxType = None,
                autoCrop = None,
                pageSize=None,
                callback=None,
                user_data=None,
                ):
        self.fileName = fileName
        self._kwds = dict(
                        pages = pages,
                        dx = dx,
                        dy = dy,
                        sx = sx,
                        sy = sy,
                        degrees = degrees,
                        orientation = orientation,
                        isdata = isdata,
                        leadingBreak = leadingBreak,
                        template = template,
                        outlineText = outlineText,
                        outlineLevel = outlineLevel,
                        outlineClosed = outlineClosed,
                        pdfBoxType = pdfBoxType,
                        autoCrop = autoCrop,
                        pageSize = pageSize,
                        callback = callback,
                        user_data = user_data,
                        )
        self._list = None

    def wrap(self,aW,aH):
        '''force a split'''
        return aW, 0x7fffffff

    def split(self,aW,aH):
        if not self._list:
            self._list = includePdfFlowables(self.fileName,**self._kwds)
        return self._list

    def identity(self, maxLen=None):
        msg = "<%s at %s%s> fileName=%s" % (self.__class__.__name__,hex(id(self)),self._frameName(),self.fileName)
        if maxLen:
            return msg[0:maxLen]
        else:
            return msg

def loadPdf(filename, canvas, pageNumbers=None, prefix=None):
    if prefix is None:
        prefix = os.path.splitext(filename)[0] + '_page'
    prefix = prefix.replace('/','_')
    pdfContent = open(filename,"rb").read()
    (formNames, stuff) = storeFormsInMemory(pdfContent,
                                            pagenumbers=pageNumbers,
                                            prefix=prefix,
                                            all=1)

    if pageNumbers:
        namesToInclude = []
        for num in pageNumbers:
            namesToInclude.append(formNames[num])
    else:
        namesToInclude = None
    restoreFormsInMemory(stuff, canvas,
                         allowDuplicates=1,
                         formnames=namesToInclude)
    return formNames



#Copyright ReportLab Europe Ltd. 2000-2017
#see license.txt for license details
import os, sys, glob, shutil
platform = sys.platform
pjoin = os.path.join
abspath = os.path.abspath
isfile = os.path.isfile
isdir = os.path.isdir
dirname = os.path.dirname
basename = os.path.basename
if __name__=='__main__':
    pkgDir=dirname(sys.argv[0])
else:
    pkgDir=dirname(__file__)
if not pkgDir:
    pkgDir=os.getcwd()
elif not os.path.isabs(pkgDir):
    pkgDir=os.path.abspath(pkgDir)

from setuptools import setup, find_packages

def get_version():
    #try source
    FN = pjoin(pkgDir,'src','rlextra','__init__')
    for l in open(pjoin(FN+'.py'),'r').readlines():
        if l.startswith('Version'):
            l = l.split('=',1)
            if len(l)==2:
                return eval(l[1].strip(),locals())
    raise ValueError('Cannot determine rlextra Version')

def main():
    setup(
        name="rlextra",
        version=get_version(),
        license="BSD license (see license.txt for details), Copyright (c) 2000-2019, ReportLab Inc.",
        description="The ReportLabPLUS Toolkit",
        long_description="""The ReportLabPLUS Toolkit. Python library for generating PDFs and graphics.""",

        author="Robinson, Watters, Lee, Precedo, Becker and many more...",
        author_email="info@reportlab.com",
        url="http://www.reportlab.com/",
        packages = find_packages("src"),
        package_dir = {'rlextra': "src/rlextra"},
        package_data = {'rlextra':[
                                'rml2pdf/00README.txt',
                                'rml2pdf/*.dtd',
                                'pageCatcher/00REAME.txt',
                                'pageCatcher/replogo.bmp',
                                'pageCatcher/testfile.pdf',
                                'security/*.txt',
                                'graphics/guiedit/sample.csv',
                                ],
                        },
        entry_points = dict(
                            console_scripts = [
                                'rml2pdf=rlextra.rml2pdf.rml2pdf:main',
                                'pagecatcher=rlextra.pageCatcher.pageCatcher:scriptInterp',
                                'pdfexplorer=rlextra.pageCatcher.pdfexplorer:test',
                                'svglib=rlextra.thirdparty.svg.svglib:main',
                                ],
                            gui_scripts = [
                                'diagra=rlextra.graphics.guiedit.guiedit:mainApp',
                                ],
                            ),
        install_requires=[
            'reportlab>=3.5.27',
            'Pmw>=2.0.0',
            'preppy>=2.7.0',
            'pyRXP>=2.1.1',
            ],
        extras_require=dict(
            digisig=[
                'asn1crypto','cffi','cryptography','enum34','ipaddress','pycryptodome',
                'pyOpenSSL','six','sortedcontainers',
                ],
            ),
        )

if __name__=='__main__':
    main()



#example9_barcode
"""
pageCatcher demo script:
   Adds a barcode onto scaled pages of another PDF file.
   Very simplistic use of barcodes for demo purposes only.
"""

import sys
from reportlab.pdfgen import canvas
try:
    from rlextra.pageCatcher.pageCatcher import storeForms, restoreForms #, fourPage
except ImportError:
    pass # running inside pageCatcher module?

from reportlab.graphics.shapes import Drawing
from reportlab.graphics.barcode.widgets import BarcodeStandard93


def run():
    inputFileName = "sample2.pdf"
    outputFileName = "out9_barcodes.pdf"
    ForegroundStorageFile = "foreground.data"

    print("placing barcodes on pages of", inputFileName)
    print("storing to", outputFileName)


    print("storing all pages of", inputFileName, "in", ForegroundStorageFile)
    storeForms(inputFileName, ForegroundStorageFile, prefix="foreground", all=1)
    print("now restoring the pages as forms")
    canv = canvas.Canvas(outputFileName)
    ForeGroundFormNames = restoreForms(ForegroundStorageFile, canv)
    pagenum = 0

    print("now writing out the pages with barcodes")
    for f in ForeGroundFormNames:
        print(f)
        pagenum += 1
        # display foreground scaled at 75%
        canv.saveState()
        canv.scale(0.75, 0.75)
        canv.doForm(f)
        canv.restoreState()

        #Create a blank Drawing.  This would commonly be
        #used for charts, but BarCode widgest can be added too
        d= Drawing(100,50)
        #create and set up the widget
        bc = BarcodeStandard93()
        bc.value = 'DEM-%06d' % (pagenum+1)
        d.add(bc) 
        d.drawOn(canv, 25, 720)

        canv.showPage()
    canv.save()
    print("wrote", outputFileName)
    

if __name__=='__main__':
    run()

    
